# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for client."""

import abc
import functools
from typing import Callable, Generic, Optional, TypeVar

from absl import logging
from vizier import pyvizier as vz
from vizier.client import client_abc

from absl.testing import parameterized

# Aliases are defined, so when you are developing a new client, you can
# swap it with your subclass. It makes your IDE understand which class
# you are using.
_S = TypeVar('_S', bound=client_abc.StudyInterface)
_TrialClient = client_abc.TrialInterface


class VizierClientTestMixin(abc.ABC, Generic[_S]):

  @abc.abstractmethod
  def create_study(self, study_config: vz.ProblemStatement, name: str) -> _S:
    """Create study given study config and study name."""
    pass


class MyMeta(type(VizierClientTestMixin), type(parameterized.TestCase)):
  pass


class TestCase(
    parameterized.TestCase, VizierClientTestMixin[_S], metaclass=MyMeta
):
  """Generic tests for cross-platform clients.

  This test provides basic coverage and it is not meant to be a comprehensive
  e2e test on all possible inputs to the client.

  Override `create_study` method.
  """

  def create_test_study(self, name: str) -> _S:
    """Creates a study."""
    logging.info('Creating study name=%s, testcasename=%s', name, self.id())
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    problem.metric_information.append(
        vz.MetricInformation(
            name='maximize_metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    study = self.create_study(problem, name)

    # TODO: Put this line back once we have __eq__ well-defined.
    # self.assertEqual(study.materialize_problem_statement(), problem)
    return study

  def test_create_or_load_study(self):
    study = self.create_test_study(self.id())
    study2 = study.from_resource_name(study.resource_name)
    self.assertEqual(study.resource_name, study2.resource_name)

    self.assertEqual(
        study.materialize_problem_statement(),
        study2.materialize_problem_statement(),
    )

  def test_delete_study(self):
    study = self.create_test_study(self.id())
    resource_name = study.resource_name
    study.delete()

    with self.assertRaises(Exception):
      study.from_resource_name(resource_name)

  def _example_trials(self) -> list[vz.Trial]:
    """Generates example trials."""
    trials = [
        # Completed trial.
        vz.Trial(
            parameters={'float': 0.5},
        ).complete(vz.Measurement({'maximize_metric': 1.0})),
        # Completed trial.
        vz.Trial(
            parameters={'float': 0.5},
        ).complete(vz.Measurement({'maximize_metric': 0.5})),
        # Requested trial, which will be made active below.
        vz.Trial(
            parameters={'float': 0.5},
            measurements=[vz.Measurement({'maximize_metric': 0.7})],
        ),
        # Requested trial.
        vz.Trial(parameters={'float': 0.5}, is_requested=True),
    ]
    for idx, t in enumerate(trials):
      t.metadata['future_id'] = str(idx + 1)  # id to be assigned
    return trials

  def create_test_study_with_trials(self, name: str) -> _S:
    study = self.create_test_study(name)
    trials = self._example_trials()
    for i, t in enumerate(trials):
      # TODO: Remove this.
      study._add_trial(t)  # pylint: disable=protected-access
      if i == 2:
        # Make sure the requested trial becomes ACTIVE.
        _ = study.suggest(count=1)
    return study

  @parameterized.parameters(list(state for state in vz.StudyState))
  def test_set_state(self, target_state: vz.StudyState):
    study = self.create_test_study_with_trials(self.id())
    # TODO: Check that the study moved to the target state.
    # This test is currently a placeholder.
    try:
      study.set_state(target_state)
    except NotImplementedError:
      logging.exception(
          'Set study state for %s is not implemented in %s',
          target_state,
          type(self),
      )

  def test_list_trials(self):
    study = self.create_test_study_with_trials(self.id())
    self.assertLen(list(study.trials()), 4)

  def test_trials_iter_and_get_are_equal(self):
    study = self.create_test_study_with_trials(self.id())
    all_trials = study.trials()
    self.assertEqual(
        [t.id for t in all_trials], [t.id for t in all_trials.get()]
    )

  def test_optimal_trials_on_empty(self):
    study = self.create_test_study(self.id())
    self.assertEmpty(list(study.optimal_trials()))

  def test_optimal_trials_notempty(self):
    study = self.create_test_study_with_trials(self.id())
    optimal_trials = list(study.optimal_trials())
    self.assertLen(optimal_trials, 1)
    self.assertEqual(
        optimal_trials[0].materialize().status, vz.TrialStatus.COMPLETED
    )

  def test_suggest_same_worker(self):
    # Given the same client id, should return the same trials.
    study = self.create_test_study(self.id())
    trials1 = study.suggest(count=5, client_id='worker1')
    self.assertLen(trials1, 5)
    trials2 = study.suggest(count=2, client_id='worker1')
    self.assertTrue({t.id for t in trials2}.issubset({t.id for t in trials1}))

  def test_suggest_requested(self):
    # Given the same client id, should return the same trials.
    study = self.create_test_study(self.id())
    requested_parameters = {'float': 0.11112}
    requested_trial = study.request(vz.TrialSuggestion(requested_parameters))
    self.assertCountEqual(
        requested_trial.parameters.items(), requested_parameters.items()
    )
    trials = study.suggest(count=5, client_id='worker1')
    self.assertLen(trials, 5)
    self.assertCountEqual(
        trials[0].parameters.items(), requested_parameters.items()
    )

  def test_suggest_different_workers(self):
    # Given different client ids, should generate new suggestions.
    study = self.create_test_study(self.id())
    worker1_trials = study.suggest(count=5, client_id='worker1')
    self.assertLen(worker1_trials, 5)
    worker1_trial_ids = {t.id for t in worker1_trials}
    worker2_trials = study.suggest(count=2, client_id='worker2')
    self.assertLen(worker2_trials, 2)
    worker2_trial_ids = {t.id for t in worker2_trials}
    self.assertEmpty(
        worker1_trial_ids.intersection(worker2_trial_ids),
        msg=(
            f'worker1_trial_ids={worker1_trial_ids}\n'
            f'worker2_trial_ids={worker2_trial_ids}'
        ),
    )

  def test_suggest_with_immutable_study(self):
    # Given immutable study, suggestions should be empty.
    study = self.create_test_study(self.id())
    study.set_state(vz.StudyState.ABORTED)
    trials = study.suggest(count=5, client_id='worker1')
    self.assertEmpty(trials)

  def test_get_trial(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(2).materialize()
    self.assertEqual(trial.metadata['future_id'], '2')
    self.assertEqual(trial.id, 2)

  def test_get_trials(self):
    study = self.create_test_study_with_trials(self.id())
    completed = study.trials(vz.TrialFilter(status={vz.TrialStatus.COMPLETED}))
    self.assertLen(list(completed), 2)

  def test_parameters(self):
    study = self.create_test_study_with_trials(self.id())
    self.assertEqual(
        study.get_trial(1).parameters, study.get_trial(1).parameters
    )

  def test_study_update_metadata(self):
    """Checks for correct merge behavior."""
    study = self.create_test_study(self.id())
    delta_metadata = vz.Metadata({'bar': 'bar_v'}, foo='foo_v')
    study.update_metadata(delta_metadata)
    self.assertEqual(
        study.materialize_problem_statement().metadata.get('bar'), 'bar_v'
    )

    delta_metadata_2 = vz.Metadata({'bar': 'bar_w'})
    study.update_metadata(delta_metadata_2)

    problem_statement = study.materialize_problem_statement()
    self.assertEqual(problem_statement.metadata.get('bar'), 'bar_w')
    self.assertEqual(problem_statement.metadata.get('foo'), 'foo_v')

  ###########################################
  ## Tests on Trial class start here.
  ###########################################

  def test_delete_trial(self):
    study = self.create_test_study_with_trials(self.id())
    study.get_trial(2).delete()
    with self.assertRaises(Exception):
      study.get_trial(2)

  def test_complete_trial_no_measurements(self):
    study = self.create_test_study_with_trials(self.id())
    with self.assertRaises(Exception):
      study.get_trial(4).complete()

  def test_complete_trial_auto_selection(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    measurement = trial.complete()
    self.assertIn(measurement, trial.materialize().measurements)

  def test_complete_trial_no_measurements_infeasible(self):
    study = self.create_test_study_with_trials(self.id())
    # Delete trial 3 so it's not suggested.
    trial = study.get_trial(3)
    trial.delete()
    # Ask Vizier to suggest the trial so it becomes ACTIVE.
    trial = study.suggest(count=1)[0]
    self.assertEqual(trial.id, 4)
    self.assertIsNone(trial.complete(infeasible_reason='just because'))
    self.assertTrue(trial.materialize().infeasible)

  def test_complete_trial_manual_measurement(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    final_measurement = vz.Measurement(metrics={'maximize_metric': 0.1})
    self.assertEqual(trial.complete(final_measurement), final_measurement)
    self.assertEqual(trial.materialize().status, vz.TrialStatus.COMPLETED)

  def test_should_trial_stop(self):
    # TODO: Improve coverage.
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    self.assertIsInstance(trial.check_early_stopping(), bool)

  def test_trial_stop(self):
    """Checks for correct stopping behavior."""
    study = self.create_test_study_with_trials(self.id())
    active_trial = study.suggest(count=1)[0]
    active_trial.stop()
    self.assertEqual(active_trial.materialize().status, vz.TrialStatus.STOPPING)

    # COMPLETED, STOPPING
    noop_trials = [study.get_trial(2), active_trial]
    original_statuses = [trial.materialize().status for trial in noop_trials]
    for trial, status in zip(noop_trials, original_statuses):
      trial.stop()  # Should do nothing.
      self.assertEqual(trial.materialize().status, status)

    requested_trial = study.get_trial(4)
    with self.assertRaises(Exception):
      requested_trial.stop()

  def test_intermediate_measurement(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    before = trial.materialize()
    trial.add_measurement(
        vz.Measurement(steps=2, metrics={'maximize_metric': 0.2})
    )
    after = trial.materialize()
    self.assertLen(after.measurements, len(before.measurements) + 1)

  def test_intermediate_measurement_infeasible(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    trial.complete(infeasible_reason='just because')
    before = trial.materialize()
    trial.add_measurement(
        vz.Measurement(steps=2, metrics={'maximize_metric': 0.2})
    )
    after = trial.materialize()
    # Can't add measurements to infeasible trials, but doing so isn't an error.
    self.assertEqual(before, after)

  def test_intermediate_measurement_completed(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    trial.complete(vz.Measurement(steps=1, metrics={'maximize_metric': 0.5}))
    # Can't add measurements to completed trials; doing so is an error.
    with self.assertRaises(Exception):
      trial.add_measurement(
          vz.Measurement(steps=2, metrics={'maximize_metric': 0.2})
      )

  def test_study_property(self):
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(2)
    self.assertEqual(
        trial.materialize(), trial.study.get_trial(2).materialize()
    )

  def assertPassesE2ETuning(
      self,
      *,
      study_factory: Optional[Callable[[vz.ProblemStatement], _S]] = None,
      batch_size: int = 2,
      num_iterations: int = 5,
      multi_objective: bool = False,
  ):
    """Runs an e2e test.

    This test simulates a hyperparameter tuning scenario, where the model
    hyperparameters are being tuned, and automated trial early stopping is
    used to stop trials early.

    Args:
      study_factory: If not specified, uses the default factory of this class.
      batch_size:
      num_iterations:
      multi_objective:
    """
    study_factory = study_factory or functools.partial(
        self.create_study, study_id=self.id()
    )

    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param(
        'learning_rate', min_value=0.0, max_value=1.0, default_value=0.5
    )
    problem.search_space.root.add_int_param(
        'num_layers', min_value=1, max_value=5
    )
    problem.metric_information = [
        vz.MetricInformation(
            name='accuracy', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    ]
    if multi_objective:
      problem.metric_information.append(
          vz.MetricInformation(
              name='latency', goal=vz.ObjectiveMetricGoal.MINIMIZE
          )
      )

    study: _S = study_factory(problem)

    def learning_curve_simulator(learning_rate: float) -> list[float]:
      return [learning_rate * step for step in range(10)]

    for _ in range(num_iterations):
      suggestions = study.suggest(count=batch_size)
      for trial in suggestions:
        trial: _TrialClient = trial
        assert isinstance(trial, _TrialClient), type(trial)
        learning_rate = trial.parameters['learning_rate']
        num_layers = trial.parameters['num_layers']
        possible_curve = learning_curve_simulator(learning_rate)
        evaluated_curve = []
        for i, obj in enumerate(possible_curve, start=1):
          if i > 1 and trial.check_early_stopping():
            break
          evaluated_curve.append(obj)
          trial.add_measurement(
              vz.Measurement(
                  steps=i,
                  elapsed_secs=0.1 * i,
                  metrics={'accuracy': obj, 'latency': 0.5 * num_layers},
              )
          )
        trial.complete()
        stored_curve = [
            m.metrics['accuracy'].value
            for m in trial.materialize().measurements
        ]

        # See if curve was stored correctly.
        self.assertEqual(evaluated_curve, stored_curve)

        # See if final_measurement is defaulted to end of curve.
        final_measurement = trial.materialize().final_measurement
        assert final_measurement is not None
        final_accuracy = final_measurement.metrics['accuracy'].value
        self.assertEqual(evaluated_curve[-1], final_accuracy)

  def test_trial_update_metadata(self):
    """Checks for correct merge behavior."""
    study = self.create_test_study_with_trials(self.id())
    trial = study.get_trial(3)
    delta_metadata = vz.Metadata({'bar': 'bar_v'}, foo='foo_v')
    trial.update_metadata(delta_metadata)
    self.assertEqual(trial.materialize().metadata.get('bar'), 'bar_v')
    delta_metadata_2 = vz.Metadata({'bar': 'bar_w'})
    trial.update_metadata(delta_metadata_2)
    self.assertEqual(trial.materialize().metadata.get('bar'), 'bar_w')
    self.assertEqual(trial.materialize().metadata.get('foo'), 'foo_v')



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.service.stubs_util."""

import grpc
from vizier._src.service import stubs_util
from absl.testing import absltest
from absl.testing import parameterized


class StubsUtilTest(parameterized.TestCase):

  @parameterized.parameters(
      (stubs_util.create_pythia_server_stub,),
      (stubs_util.create_vizier_server_stub,),
  )
  def test_bad_endpoint(self, stub_creator):
    # Make sure connection errors out instead of stalling forever.
    endpoint = 'i_dont_exist'
    with self.assertRaises(grpc.FutureTimeoutError):
      stub_creator(endpoint, timeout=0.1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Below contain utilities for writing tests to reduce code duplication."""
from typing import List, Sequence

from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_oss_pb2

from google.longrunning import operations_pb2


def generate_study(owner_id: str = 'my_username',
                   study_id: str = '1234',
                   display_name: str = 'cifar10',
                   **study_kwargs) -> study_pb2.Study:
  study_name = resources.StudyResource(owner_id, study_id).name
  return study_pb2.Study(
      name=study_name, display_name=display_name, **study_kwargs)


def generate_trials(trial_id_list: Sequence[int],
                    owner_id: str = 'my_username',
                    study_id: str = '1234',
                    **trial_kwargs) -> List[study_pb2.Trial]:
  """Generates arbitrary trials."""
  trials = []
  for trial_id in trial_id_list:
    trial = study_pb2.Trial(
        name=resources.TrialResource(owner_id, study_id, trial_id).name,
        id=str(trial_id),
        **trial_kwargs)
    trials.append(trial)
  return trials


def generate_all_states_trials(start_trial_index: int,
                               owner_id: str = 'my_username',
                               study_id: str = '1234',
                               **trial_kwargs) -> List[study_pb2.Trial]:
  """Generates a trial for each possible trial state."""
  trials = []
  for i, state in enumerate(study_pb2.Trial.State.keys()):
    trial_id = start_trial_index + i
    trial = study_pb2.Trial(
        name=resources.TrialResource(owner_id, study_id, trial_id).name,
        id=str(trial_id),
        state=state,
        **trial_kwargs)
    trials.append(trial)
  return trials


def generate_suggestion_operations(
    operation_numbers: Sequence[int],
    owner_id: str = 'my_username',
    study_id: str = 'cifar10',
    client_id: str = 'client0',
    **operation_kwargs) -> List[operations_pb2.Operation]:
  """Generates arbitrary suggestion operations."""
  operations = []
  for operation_number in operation_numbers:
    operation = operations_pb2.Operation(
        name=resources.SuggestionOperationResource(owner_id, study_id,
                                                   client_id,
                                                   operation_number).name,
        **operation_kwargs)
    operations.append(operation)
  return operations


def generate_early_stopping_operations(
    trial_id_list: Sequence[int],
    owner_id: str = 'my_username',
    study_id: str = '1234',
    **operation_kwargs) -> List[vizier_oss_pb2.EarlyStoppingOperation]:
  """Generates arbitrary early stopping operations."""
  operations = []
  for trial_id in trial_id_list:
    operation = vizier_oss_pb2.EarlyStoppingOperation(
        name=resources.EarlyStoppingOperationResource(owner_id, study_id,
                                                      trial_id).name,
        **operation_kwargs)
    operations.append(operation)
  return operations


def generate_all_four_parameter_specs(**study_spec_kwargs
                                     ) -> study_pb2.StudySpec:
  """All possible primitive parameter specs for testing."""
  double_value_spec = study_pb2.StudySpec.ParameterSpec.DoubleValueSpec(
      min_value=-1.0, max_value=1.0)
  double_parameter_spec = study_pb2.StudySpec.ParameterSpec(
      parameter_id='learning_rate', double_value_spec=double_value_spec)

  integer_value_spec = study_pb2.StudySpec.ParameterSpec.IntegerValueSpec(
      min_value=1, max_value=10)
  integer_parameter_spec = study_pb2.StudySpec.ParameterSpec(
      parameter_id='num_layers', integer_value_spec=integer_value_spec)

  categorical_value_spec = (
      study_pb2.StudySpec.ParameterSpec.CategoricalValueSpec(
          values=['relu', 'sigmoid']
      )
  )
  categorical_parameter_spec = study_pb2.StudySpec.ParameterSpec(
      parameter_id='nonlinearity',
      categorical_value_spec=categorical_value_spec)

  discrete_value_spec = study_pb2.StudySpec.ParameterSpec.DiscreteValueSpec(
      values=[1.0, 1.5, 3.0, 4.5])
  discrete_parameter_spec = study_pb2.StudySpec.ParameterSpec(
      parameter_id='discrete_unnamed', discrete_value_spec=discrete_value_spec)

  return study_pb2.StudySpec(
      parameters=[
          double_parameter_spec, integer_parameter_spec,
          categorical_parameter_spec, discrete_parameter_spec
      ],
      **study_spec_kwargs)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from absl.testing import parameterized
import jax
from jax import config
import numpy as np
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import multitask_tuned_gp_models
from vizier.jax import optimizers

from absl.testing import absltest


class MultitaskTunedGpModelsTest(parameterized.TestCase):

  def _generate_xys(self):
    # x_obs has shape [10, 6] and y_obs [10, 2]
    x_obs = np.array(
        [
            [
                0.2941264,
                0.29313548,
                0.68817519,
                0.37502566,
                0.48356813,
                0.34127283,
            ],
            [
                0.66218224,
                0.70770083,
                0.6901334,
                0.66787973,
                0.5400858,
                0.52721233,
            ],
            [
                0.88469647,
                0.50593371,
                0.83160862,
                0.58674892,
                0.42145673,
                0.31749428,
            ],
            [
                0.39976682,
                0.59517741,
                0.73295106,
                0.6084903,
                0.54891015,
                0.44338632,
            ],
            [
                0.8354305,
                0.87605574,
                0.47855956,
                0.48174861,
                0.37685449,
                0.38348768,
            ],
            [
                0.55608455,
                0.72781129,
                0.52432913,
                0.44291417,
                0.3816395,
                0.326599,
            ],
            [
                0.24689187,
                0.50979672,
                0.67604857,
                0.45172594,
                0.34994392,
                0.75239792,
            ],
            [
                0.71007257,
                0.60896354,
                0.29270877,
                0.74683367,
                0.50169051,
                0.74480515,
            ],
            [
                0.9193235,
                0.24393112,
                0.63868591,
                0.43271524,
                0.43339578,
                0.59413154,
            ],
            [
                0.51850627,
                0.62689204,
                0.76134879,
                0.65990021,
                0.82350868,
                0.7429215,
            ],
        ],
        dtype=np.float64,
    )
    y_obs = np.transpose(
        np.array(
            [
                [
                    0.55552674,
                    -0.29054829,
                    -0.04703586,
                    0.0217839,
                    0.15445438,
                    0.46654119,
                    0.12255823,
                    -0.19540335,
                    -0.11772564,
                    -0.44447326,
                ],
                [
                    1.55552674,
                    -1.29054829,
                    -1.04703586,
                    1.0217839,
                    1.15445438,
                    1.46654119,
                    1.12255823,
                    -1.19540335,
                    -1.11772564,
                    -1.44447326,
                ],
            ],
            dtype=np.float64,
        )
    )
    return x_obs, y_obs

  @parameterized.parameters(
      multitask_tuned_gp_models.MultiTaskType.INDEPENDENT,
      multitask_tuned_gp_models.MultiTaskType.SEPARABLE_NORMAL_TASK_KERNEL_PRIOR,
      multitask_tuned_gp_models.MultiTaskType.SEPARABLE_LKJ_TASK_KERNEL_PRIOR,
      multitask_tuned_gp_models.MultiTaskType.SEPARABLE_DIAG_TASK_KERNEL_PRIOR,
  )
  def test_masking_works(self, multitask_type):
    # Mask three dimensions and four observations.
    x_obs, y_obs = self._generate_xys()
    categorical_dim = 3
    continuous_dim = 9
    data = types.ModelData(
        features=types.ModelInput(
            continuous=types.PaddedArray.from_array(
                x_obs, target_shape=(10, continuous_dim), fill_value=1.0
            ),
            categorical=types.PaddedArray.from_array(
                np.zeros((9, 0), dtype=types.INT_DTYPE),
                target_shape=(10, categorical_dim),
                fill_value=1,
            ),
        ),
        # y_obs has shape [10, 2]
        labels=types.PaddedArray.as_padded(y_obs),
    )
    model1 = sp.CoroutineWithData(
        multitask_tuned_gp_models.VizierMultitaskGaussianProcess(
            _feature_dim=types.ContinuousAndCategorical[int](
                continuous_dim, categorical_dim
            ),
            _num_tasks=2,
            _multitask_type=multitask_type,
        ),
        data=data,
    )

    modified_data = types.ModelData(
        features=types.ModelInput(
            continuous=data.features.continuous.replace_fill_value(np.nan),
            categorical=data.features.categorical.replace_fill_value(-1),
        ),
        labels=data.labels,
    )
    model2 = sp.CoroutineWithData(
        multitask_tuned_gp_models.VizierMultitaskGaussianProcess(
            _feature_dim=types.ContinuousAndCategorical[int](
                continuous_dim, categorical_dim
            ),
            _num_tasks=2,
            _multitask_type=multitask_type,
        ),
        data=modified_data,
    )

    # Check that the model loss and optimal parameters are independent of those
    # dimensions and observations.
    optimize = optimizers.JaxoptScipyLbfgsB()
    rng, init_rng = jax.random.split(jax.random.PRNGKey(2), 2)
    optimal_params1, _ = optimize(
        init_params=jax.vmap(model1.setup)(jax.random.split(init_rng, 1)),
        loss_fn=model1.loss_with_aux,
        rng=rng,
        constraints=sp.get_constraints(model1),
    )
    optimal_params2, _ = optimize(
        init_params=jax.vmap(model2.setup)(jax.random.split(init_rng, 1)),
        loss_fn=model2.loss_with_aux,
        rng=rng,
        constraints=sp.get_constraints(model2),
    )

    for key in optimal_params1:
      self.assertTrue(
          np.all(np.equal(optimal_params1[key], optimal_params2[key])),
          msg=f'{key} parameters were not equal.',
      )
    self.assertEqual(
        model1.loss_with_aux(optimal_params1)[0],
        model2.loss_with_aux(optimal_params2)[0],
    )


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters import switch_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob
from absl.testing import absltest


class SwitchExperimenterTest(absltest.TestCase):

  def test_switch_experimenter(self):
    exptr0 = numpy_experimenter.NumpyExperimenter(
        lambda x: 0.0 * x, bbob.DefaultBBOBProblemStatement(1, metric_name='0')
    )
    exptr1 = numpy_experimenter.NumpyExperimenter(
        lambda x: 1.0 * x, bbob.DefaultBBOBProblemStatement(1, metric_name='1')
    )
    switch_exptr = switch_experimenter.SwitchExperimenter([exptr0, exptr1])

    t0 = vz.Trial(parameters={'switch': 0, 'x0': 100.0})
    t1 = vz.Trial(parameters={'switch': 1, 'x0': 100.0})

    switch_exptr.evaluate([t0, t1])

    self.assertEqual(t0.final_measurement.metrics['switch_metric'].value, 0.0)  # pytype:disable=attribute-error
    self.assertEqual(t1.final_measurement.metrics['switch_metric'].value, 100.0)  # pytype:disable=attribute-error


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Test for VizierSearch. Cannot be tested internally but can be on GitHub."""
from ray import tune
from vizier._src.raytune import converters
from vizier._src.raytune import run_tune
from vizier._src.raytune import vizier_search
from vizier.benchmarks import experimenters
from vizier.service import clients
from vizier.service import pyvizier as vz

from absl.testing import absltest


# Required to use local Vizier service.
clients.environment_variables.servicer_use_sql_ram()


class VizierSearchTest(absltest.TestCase):

  def test_search_with_study_config(self):
    dim = 4
    bbob_factory = experimenters.BBOBExperimenterFactory(name='Sphere', dim=dim)
    exptr = bbob_factory()

    study_config = vz.StudyConfig.from_problem(exptr.problem_statement())
    self.assertLen(study_config.search_space.parameters, dim)
    searcher = vizier_search.VizierSearch(
        'test study', study_config, algorithm='RANDOM_SEARCH'
    )

    trainable = converters.ExperimenterConverter.to_callable(exptr)
    tuner = tune.Tuner(
        trainable,
        param_space=None,
        tune_config=tune.TuneConfig(num_samples=5, search_alg=searcher),
    )
    tuner.fit()
    self.assertLen(tuner.get_results(), 5)

  # def test_search_with_ray_search_space(self):
  #  dim = 4
  #  bbob_factory = experimenters.BBOBExperimenterFactory(name='Sphere',
  # dim=dim)
  #  exptr = bbob_factory()
  #
  #  trainable = converters.ExperimenterConverter.to_callable(exptr)

  #  tuner = tune.Tuner(
  #      trainable,
  #      param_space=None,
  #      tune_config=tune.TuneConfig(
  #          num_samples=10, search_alg=vizier_search.VizierSearch(),
  # max_concurrent_trials=1),
  #  )
  #  tuner.fit()
  #  self.assertLen(tuner.get_results(), 10)

  def test_random_search_with_run_tune(self):
    results = run_tune.run_tune_bbob(
        function_name='Sphere',
        dimension=3,
        shift=None,
        tune_config=tune.TuneConfig(
            search_alg=vizier_search.VizierSearch(algorithm='RANDOM_SEARCH'),
            num_samples=9,
            max_concurrent_trials=1,
        ),
        run_config=None,
    )
    self.assertLen(results, 9)

  def test_vizier_search_with_run_tune(self):
    # Use the default algorithm = GP Bandit.
    results = run_tune.run_tune_bbob(
        function_name='Sphere',
        dimension=3,
        shift=None,
        tune_config=tune.TuneConfig(
            search_alg=vizier_search.VizierSearch(),
            num_samples=5,
            max_concurrent_trials=1,
        ),
        run_config=None,
    )
    self.assertLen(results, 5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Library functions for testing databases."""
import copy
import random
import string
from typing import List

from vizier._src.service import custom_errors
from vizier._src.service import datastore
from vizier._src.service import key_value_pb2
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_oss_pb2
from vizier._src.service import vizier_service_pb2

from google.longrunning import operations_pb2
from absl.testing import parameterized

UnitMetadataUpdate = vizier_service_pb2.UnitMetadataUpdate


def make_random_string() -> str:
  return ''.join(random.choice(string.ascii_lowercase) for _ in range(10))


class DataStoreTestCase(parameterized.TestCase):
  """Base class for testing datastores."""

  def assertStudyAPI(self, ds: datastore.DataStore, study: study_pb2.Study):
    """Tests if the datastore handles studies correctly."""
    ds.create_study(study)
    with self.assertRaises(custom_errors.AlreadyExistsError):
      ds.create_study(study)  # Can't make another study w/ same name.

    copied_study = ds.load_study(study.name)
    with self.assertRaises(custom_errors.NotFoundError):
      ds.load_study(study.name + 'does_not_exist')  # Non-existent study.
    self.assertEqual(copied_study, study)
    self.assertIsNot(copied_study, study)  # Check pass-by-value.

    owner_name = resources.StudyResource.from_name(
        study.name
    ).owner_resource.name
    list_of_one_study = ds.list_studies(owner_name)
    with self.assertRaises(custom_errors.NotFoundError):
      ds.list_studies(owner_name + 'does_not_exist')  # Non-existent study.
    self.assertLen(list_of_one_study, 1)
    self.assertEqual(list_of_one_study[0], study)
    self.assertIsNot(list_of_one_study[0], study)  # Check pass-by-value.

    study.state = study_pb2.Study.State.COMPLETED
    ds.update_study(study)
    self.assertEqual(ds.load_study(study.name), study)
    self.assertIsNot(ds.load_study(study.name), study)  # Check pass-by-value.
    with self.assertRaises(custom_errors.NotFoundError):
      missing_study = copy.deepcopy(study)
      missing_study.name += str('i_dont_exist')
      ds.update_study(missing_study)  # Does not exist.

    study.inactive_reason = make_random_string()
    copied_original_study = ds.load_study(study.name)
    self.assertNotEqual(study, copied_original_study)  # Check pass-by-value.

    ds.delete_study(study.name)
    with self.assertRaises(custom_errors.NotFoundError):
      ds.delete_study(study.name)  # Deleting non-existent study.
    with self.assertRaises(custom_errors.NotFoundError):
      ds.load_study(study.name)  # Non-existent study.

    # Owner should be kept track of.
    expected_studies = ds.list_studies(owner_name)
    self.assertEmpty(expected_studies)

  def assertTrialAPI(
      self,
      ds: datastore.DataStore,
      study: study_pb2.Study,
      trials: List[study_pb2.Trial],
  ):
    """Tests if the datastore handles trials correctly."""
    num_trials = len(trials)

    ds.create_study(study)
    empty_list = ds.list_trials(study.name)
    self.assertEmpty(empty_list)

    expected_max_trial_id = ds.max_trial_id(study.name)
    self.assertEqual(expected_max_trial_id, 0)

    for trial in trials:
      ds.create_trial(trial)
      with self.assertRaises(custom_errors.AlreadyExistsError):
        ds.create_trial(trial)  # Already exists.
      copied_trial = ds.get_trial(trial.name)
      with self.assertRaises(custom_errors.NotFoundError):
        ds.get_trial(trial.name + str(num_trials))  # Does not exist.
      self.assertEqual(trial, copied_trial)
      self.assertIsNot(trial, copied_trial)  # Check pass-by-value.

    self.assertLen(trials, ds.max_trial_id(study.name))
    with self.assertRaises(custom_errors.NotFoundError):
      ds.max_trial_id(study.name + 'does_not_exist')  # Does not exist.

    list_of_trials = ds.list_trials(study.name)
    with self.assertRaises(custom_errors.NotFoundError):
      ds.list_trials(study.name + 'does_not_exist')  # Does not exist.
    self.assertEqual(list_of_trials, trials)
    self.assertIsNot(list_of_trials, trials)  # Check pass-by-value.

    first_trial = trials[0]
    first_trial.infeasible_reason = make_random_string()
    ds.update_trial(first_trial)
    with self.assertRaises(custom_errors.NotFoundError):
      missing_trial = copy.deepcopy(first_trial)
      missing_trial.name += str(num_trials)
      ds.update_trial(missing_trial)  # Does not exist.
    new_first_trial = ds.get_trial(first_trial.name)
    self.assertEqual(first_trial, new_first_trial)
    self.assertIsNot(first_trial, new_first_trial)  # Check pass-by-value.

    ds.delete_trial(first_trial.name)
    with self.assertRaises(custom_errors.NotFoundError):
      ds.delete_trial(first_trial.name)  # Already deleted.
      ds.delete_trial(first_trial.name + str(num_trials))  # Does not exist.
    leftover_trials = ds.list_trials(study.name)
    self.assertEqual(leftover_trials, trials[1:])
    self.assertIsNot(leftover_trials, trials[1:])  # Check pass-by-value.

  def assertSuggestOpAPI(
      self,
      ds: datastore.DataStore,
      study: study_pb2.Study,
      client_id: str,
      suggestion_ops: List[operations_pb2.Operation],
  ):
    """Tests if the datastore handles suggest ops correctly."""
    study_resource = resources.StudyResource.from_name(study.name)

    ds.create_study(study)
    for operation in suggestion_ops:
      ds.create_suggestion_operation(operation)
      with self.assertRaises(custom_errors.AlreadyExistsError):
        ds.create_suggestion_operation(operation)  # Already exists.

    self.assertLen(
        suggestion_ops,
        ds.max_suggestion_operation_number(study_resource.name, client_id),
    )

    with self.assertRaises(custom_errors.NotFoundError):
      ds.max_suggestion_operation_number(
          study_resource.name, client_id + 'does_not_exist'
      )  # Client doesn't exist.

    list_of_operations = ds.list_suggestion_operations(
        study_resource.name, client_id
    )
    self.assertEqual(list_of_operations, suggestion_ops)
    with self.assertRaises(custom_errors.NotFoundError):
      ds.list_suggestion_operations(
          study_resource.name, client_id + 'does_not_exist'
      )  # Client doesn't exist.

    output_op = ds.get_suggestion_operation(
        resources.SuggestionOperationResource(
            study_resource.owner_id,
            study_resource.study_id,
            client_id,
            operation_number=1,
        ).name
    )
    self.assertEqual(output_op, suggestion_ops[0])
    self.assertIsNot(output_op, suggestion_ops[0])  # Check pass-by-value.

    with self.assertRaises(custom_errors.NotFoundError):
      ds.get_suggestion_operation(
          resources.SuggestionOperationResource(
              study_resource.owner_id,
              study_resource.study_id,
              client_id + 'does_not_exist',  # Client doesn't exist.
              operation_number=1,
          ).name
      )

    output_op.metadata.type_url = make_random_string()
    ds.update_suggestion_operation(output_op)
    new_output_op = ds.get_suggestion_operation(output_op.name)
    self.assertEqual(output_op, new_output_op)

    wrong_output_op = copy.deepcopy(output_op)
    wrong_output_op.name = resources.SuggestionOperationResource(
        study_resource.owner_id,
        study_resource.study_id,
        client_id + 'does_not_exist',  # Client doesn't exist.
        operation_number=1,
    ).name
    with self.assertRaises(custom_errors.NotFoundError):
      ds.update_suggestion_operation(wrong_output_op)

  def assertEarlyStoppingAPI(
      self,
      ds: datastore.DataStore,
      study: study_pb2.Study,
      trials: List[study_pb2.Trial],
      early_stopping_ops: List[vizier_oss_pb2.EarlyStoppingOperation],
  ):
    """Tests if the datastore handles early stopping ops correctly."""
    study_resource = resources.StudyResource.from_name(study.name)
    ds.create_study(study)

    for trial in trials:
      ds.create_trial(trial)

    for operation in early_stopping_ops:
      ds.create_early_stopping_operation(operation)
      with self.assertRaises(custom_errors.AlreadyExistsError):
        ds.create_early_stopping_operation(operation)  # Op already exists.

    output_op = ds.get_early_stopping_operation(
        resources.EarlyStoppingOperationResource(
            study_resource.owner_id, study_resource.study_id, 1
        ).name
    )
    self.assertEqual(output_op, early_stopping_ops[0])
    self.assertIsNot(output_op, early_stopping_ops[0])  # Check pass-by-value.

    wrong_op_name = resources.EarlyStoppingOperationResource(
        study_resource.owner_id,
        study_resource.study_id + 'does_not_exist',  # Study doesn't exist.
        1,
    ).name
    with self.assertRaises(custom_errors.NotFoundError):
      ds.get_early_stopping_operation(wrong_op_name)

    output_op.failure_message = make_random_string()
    ds.update_early_stopping_operation(output_op)
    new_output_op = ds.get_early_stopping_operation(output_op.name)
    self.assertEqual(output_op, new_output_op)

    wrong_output_op = copy.deepcopy(output_op)
    wrong_output_op.name = resources.EarlyStoppingOperationResource(
        study_resource.owner_id,
        study_resource.study_id + 'does_not_exist',  # Study doesn't exist.
        1,
    ).name
    with self.assertRaises(custom_errors.NotFoundError):
      ds.update_early_stopping_operation(wrong_output_op)

  def assertUpdateMetadataAPI(
      self,
      ds: datastore.DataStore,
      study: study_pb2.Study,
      trials: List[study_pb2.Trial],
  ):
    """Tests if the datastore handles metadata updates properly."""
    #   We start with two metadata items in the Trial:  'd':'D' and 'e':'E',
    #   and we merge in two metadata items in $new_metadata:
    #   {'d':'Dnew' and 'f': 'Fnew'}.  The result should overwrite 'd' and
    #   be {'d':'Dnew', 'e':'E', 'f':'Fnew'}.
    study.study_spec.metadata.add(ns='s', key='d', value='D')
    study.study_spec.metadata.add(ns='s', key='e', value='E')
    ds.create_study(study)
    for trial in trials:
      trial.metadata.add(ns='t', key='d', value='D')
      trial.metadata.add(ns='t', key='e', value='E')
      ds.create_trial(trial)
    study_metadata = [
        key_value_pb2.KeyValue(key='d', ns='s', value='Dnew'),
        key_value_pb2.KeyValue(key='f', ns='s', value='Fnew'),
    ]
    trial_metadata = [
        UnitMetadataUpdate(
            trial_id='1',
            metadatum=key_value_pb2.KeyValue(key='d', ns='t', value='Dnew'),
        ),
        UnitMetadataUpdate(
            trial_id='1',
            metadatum=key_value_pb2.KeyValue(key='f', ns='t', value='Fnew'),
        ),
    ]
    ds.update_metadata(study.name, study_metadata, trial_metadata)
    mutated_study_config = ds.load_study(study.name).study_spec
    self.assertEqual(
        list(mutated_study_config.metadata),
        [
            key_value_pb2.KeyValue(ns='s', key='d', value='Dnew'),
            key_value_pb2.KeyValue(ns='s', key='e', value='E'),
            key_value_pb2.KeyValue(ns='s', key='f', value='Fnew'),
        ],
    )
    mutated_trial = ds.get_trial(trials[0].name)
    self.assertEqual(mutated_trial.id, str(trial_metadata[0].trial_id))
    self.assertEqual(
        list(mutated_trial.metadata),
        [
            key_value_pb2.KeyValue(ns='t', key='d', value='Dnew'),
            key_value_pb2.KeyValue(ns='t', key='e', value='E'),
            key_value_pb2.KeyValue(ns='t', key='f', value='Fnew'),
        ],
    )

    study_not_exist_name = study.name + 'i_dont_exist'
    with self.assertRaises(custom_errors.NotFoundError):
      ds.update_metadata(study_not_exist_name, [], [])



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import itertools
import json
import numpy as np
from vizier import benchmarks as vzb
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import grid
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.analyzers import state_analyzer
from vizier.benchmarks import experimenters
from absl.testing import absltest


class StateAnalyzerTest(absltest.TestCase):

  def test_curve_conversion(self):
    dim = 10
    experimenter = experimenters.BBOBExperimenterFactory('Sphere', dim)()

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    benchmark_state_factory = vzb.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=experimenter
    )
    num_trials = 20
    runner = vzb.BenchmarkRunner(
        benchmark_subroutines=[vzb.GenerateAndEvaluate()],
        num_repeats=num_trials,
    )

    states = []
    num_repeats = 3
    for i in range(num_repeats):
      bench_state = benchmark_state_factory(seed=i)
      runner.run(bench_state)
      states.append(bench_state)

    curve = state_analyzer.BenchmarkStateAnalyzer.to_curve(states)
    self.assertEqual(curve.ys.shape, (num_repeats, num_trials))

  def test_empty_curve_error(self):
    with self.assertRaisesRegex(ValueError, 'Empty'):
      state_analyzer.BenchmarkStateAnalyzer.to_curve([])

  def test_multiobj_curve_conversion(self):
    dim = 10
    experimenter_factories = {
        'sphere': experimenters.BBOBExperimenterFactory('Sphere', dim),
        'discus': experimenters.BBOBExperimenterFactory('Discus', dim),
    }
    multi_experimenter = experimenters.CombinedExperimenterFactory(
        base_factories=experimenter_factories
    )()

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    benchmark_state_factory = vzb.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=multi_experimenter
    )
    num_trials = 20
    runner = vzb.BenchmarkRunner(
        benchmark_subroutines=[vzb.GenerateAndEvaluate()],
        num_repeats=num_trials,
    )

    states = []
    num_repeats = 3
    for i in range(num_repeats):
      bench_state = benchmark_state_factory(seed=i)
      runner.run(bench_state)
      states.append(bench_state)

    curve = state_analyzer.BenchmarkStateAnalyzer.to_curve(
        states, reference_value=np.asarray([-1])
    )
    self.assertEqual(curve.ys.shape, (num_repeats, num_trials))

  def test_different_curve_error(self):
    exp1 = experimenters.BBOBExperimenterFactory('Sphere', dim=2)()
    exp2 = experimenters.BBOBExperimenterFactory('Sphere', dim=3)()

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    state1_factory = vzb.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=exp1
    )
    state2_factory = vzb.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=exp2
    )

    runner = vzb.BenchmarkRunner(
        benchmark_subroutines=[vzb.GenerateAndEvaluate()],
        num_repeats=10,
    )

    state1 = state1_factory()
    state2 = state2_factory()
    runner.run(state1)
    runner.run(state2)

    with self.assertRaisesRegex(ValueError, 'must have same problem'):
      state_analyzer.BenchmarkStateAnalyzer.to_curve([state1, state2])

  def test_record_conversion(self):
    dim = 10
    factory = experimenters.BBOBExperimenterFactory('Sphere', dim)
    experimenter = factory()

    def _designer_factory(config: vz.ProblemStatement, seed: int):
      return random.RandomDesigner(config.search_space, seed=seed)

    benchmark_state_factory = vzb.DesignerBenchmarkStateFactory(
        designer_factory=_designer_factory, experimenter=experimenter
    )
    num_trials = 20
    runner = vzb.BenchmarkRunner(
        benchmark_subroutines=[vzb.GenerateAndEvaluate()],
        num_repeats=num_trials,
    )

    states = []
    num_repeats = 3
    for i in range(num_repeats):
      bench_state = benchmark_state_factory(seed=i)
      runner.run(bench_state)
      states.append(bench_state)

    record = state_analyzer.BenchmarkStateAnalyzer.to_record(
        algorithm='random', experimenter_factory=factory, states=states
    )
    objective_curve = record.plot_elements['objective'].curve
    self.assertIsNotNone(objective_curve)
    self.assertEqual(objective_curve.ys.shape, (num_repeats, num_trials))
    self.assertEqual(record.algorithm, 'random')
    self.assertIn('Sphere', str(record.experimenter_metadata))
    self.assertIn(f'{dim}', str(record.experimenter_metadata))

  def test_summarize_records(self):
    function_names = ['Sphere', 'Discus', 'AttractiveSector']
    dimensions = [4, 8]
    product_list = list(itertools.product(function_names, dimensions))

    experimenter_factories = []
    for product in product_list:
      experimenter_factory = experimenters.BBOBExperimenterFactory(
          name=product[0], dim=product[1]
      )
      experimenter_factories.append(experimenter_factory)

    num_repeats = 5
    num_iterations = 20
    runner = vzb.BenchmarkRunner(
        benchmark_subroutines=[
            vzb.GenerateSuggestions(),
            vzb.EvaluateActiveTrials(),
        ],
        num_repeats=num_iterations,
    )
    algorithms = {
        'grid': grid.GridSearchDesigner.from_problem,
        'random': random.RandomDesigner.from_problem,
    }

    records = []
    for experimenter_factory in experimenter_factories:
      for algo_name, algo_factory in algorithms.items():
        benchmark_state_factory = vzb.ExperimenterDesignerBenchmarkStateFactory(
            experimenter_factory=experimenter_factory,
            designer_factory=algo_factory,
        )
        states = []
        for _ in range(num_repeats):
          benchmark_state = benchmark_state_factory()
          runner.run(benchmark_state)
          states.append(benchmark_state)
        record = state_analyzer.BenchmarkStateAnalyzer.to_record(
            algorithm=algo_name,
            experimenter_factory=experimenter_factory,
            states=states,
        )
        records.append(record)

    def record_to_reduced_keys(record: state_analyzer.BenchmarkRecord):
      json_metadata = record.experimenter_metadata[
          experimenters.experimenter_factory.BBOB_FACTORY_KEY
      ]
      metadata_dict = json.loads(json_metadata)
      exptr_name = metadata_dict.pop('name')
      return (exptr_name, json.dumps(metadata_dict))

    summarized_records = state_analyzer.BenchmarkRecordAnalyzer.summarize(
        records, record_to_reduced_keys
    )
    self.assertLen(summarized_records, len(dimensions) * len(algorithms))
    for record in summarized_records:
      self.assertLen(record.plot_elements, 1)


if __name__ == '__main__':
  absltest.main()
